@@ -1,24 +1,41 @@
-# -*- coding: UTF-8 -*-
+# -*- coding: utf-8 -*-
 from __future__ import unicode_literals, print_function
-import astroid
-import six
+
+import ast
+import inspect
+import io
+import os
+import re
 import sys
 import textwrap
 import unittest
+from time import time
+
+import astroid
+import six
+from asttokens import util, ASTTokens
+
 from . import tools
 
 
 class TestMarkTokens(unittest.TestCase):
+  maxDiff = None
 
   # We use the same test cases to test both nodes produced by the built-in `ast` module, and by
   # the `astroid` library. The latter derives TestAstroid class from TestMarkTokens. For checks
   # that differ between them, .is_astroid_test allows to distinguish.
   is_astroid_test = False
+  module = ast
 
-  @classmethod
-  def create_mark_checker(cls, source):
-    return tools.MarkChecker(source, parse=True)
+  def create_mark_checker(self, source, verify=True):
+    checker = tools.MarkChecker(self.create_asttokens(source))
+    if verify:
+      checker.verify_all_nodes(self)
+    return checker
 
+  @staticmethod
+  def create_asttokens(source):
+    return ASTTokens(source, parse=True)
 
   def print_timing(self):
     # Print the timing of mark_tokens(). This doesn't normally run as a unittest, but if you'd like
@@ -97,13 +114,13 @@ b +     # line3
       # All other expressions preserve newlines and comments but are parenthesized.
       'b +     # line3\n  c',
       'b +     # line3\n  c +   # line4\n  d',
-      'a,      # line2\nb +     # line3\n  c +   # line4\n  d',
     })
+    self.assertIn('Tuple:' + source, m.view_nodes_at(1, 0))
 
 
   def verify_fixture_file(self, path):
     source = tools.read_fixture(path)
-    m = self.create_mark_checker(source)
+    m = self.create_mark_checker(source, verify=False)
     tested_nodes = m.verify_all_nodes(self)
 
     exp_index = (0 if six.PY2 else 1) + (2 if self.is_astroid_test else 0)
@@ -124,7 +141,7 @@ b +     # line3
     'astroid/email.py':               ( 3,    3,    1,    1,   ),
     'astroid/format.py':              ( 64,   61,   62,   62,  ),
     'astroid/module.py':              ( 185,  174,  171,  171, ),
-    'astroid/module2.py':             ( 248,  253,  235,  248, ),
+    'astroid/module2.py':             ( 248,  253,  240,  253, ),
     'astroid/noendingnewline.py':     ( 57,   59,   57,   63,  ),
     'astroid/notall.py':              ( 15,   17,   15,   17,  ),
     'astroid/recursion.py':           ( 6,    6,    4,    4,   ),
@@ -155,13 +172,19 @@ b +     # line3
     # to_source() on it because it chokes on recursion depth. So we test individual nodes.
     source = tools.read_fixture('astroid/joined_strings.py')
 
-    astroid.MANAGER.optimize_ast = True
-    try:
-      m = self.create_mark_checker(source)
-    finally:
-      astroid.MANAGER.optimize_ast = False
-
     if self.is_astroid_test:
+      if getattr(astroid, '__version__', '1') >= '2':
+        # Astroid 2 no longer supports this; see
+        # https://github.com/PyCQA/astroid/issues/557#issuecomment-396004274
+        self.skipTest('astroid-2.0 does not support this')
+
+      # Astroid < 2 does support this with optimize_ast set to True
+      astroid.MANAGER.optimize_ast = True
+      try:
+        m = self.create_mark_checker(source, verify=False)
+      finally:
+        astroid.MANAGER.optimize_ast = False
+
       self.assertEqual(len(m.all_nodes), 4)     # This is the result of astroid's optimization
       self.assertEqual(m.view_node_types_at(1, 0), {'Module', 'Assign', 'AssignName'})
       const = next(n for n in m.all_nodes if isinstance(n, astroid.nodes.Const))
@@ -171,13 +194,14 @@ b +     # line3
       # astroid could avoid the need for the optimization by using an explicit stack like we do.
       #self.assertEqual(m.atok.get_text_range(const), (5, len(source) - 1))
     else:
+      m = self.create_mark_checker(source, verify=False)
       self.assertEqual(len(m.all_nodes), 2104)
       self.assertEqual(m.view_node(m.all_nodes[-1]),
-                       "Str:'F1akOFFiRIgPHTZksKBAgMCLGTdGNIAAQgKfDAcgZbj0odOnUA8GBAA7'")
+                       "Constant:'F1akOFFiRIgPHTZksKBAgMCLGTdGNIAAQgKfDAcgZbj0odOnUA8GBAA7'")
       self.assertEqual(m.view_node(m.all_nodes[-2]),
-                       "Str:'Ii0uLDAxLzI0Mh44U0gxMDI5JkM0JjU3NDY6Kjc5Njo7OUE8Ozw+Oz89QTxA'")
+                       "Constant:'Ii0uLDAxLzI0Mh44U0gxMDI5JkM0JjU3NDY6Kjc5Njo7OUE8Ozw+Oz89QTxA'")
       self.assertEqual(m.view_node(m.all_nodes[1053]),
-                       "Str:'R0lGODlhigJnAef/AAABAAEEAAkCAAMGAg0GBAYJBQoMCBMODQ4QDRITEBkS'")
+                       "Constant:'R0lGODlhigJnAef/AAABAAEEAAkCAAMGAg0GBAYJBQoMCBMODQ4QDRITEBkS'")
       self.assertEqual(m.view_node(m.all_nodes[1052]),
                        "BinOp:'R0lGODlhigJnAef/AAABAAEEAAkCAAMGAg0GBAYJBQoMCBMODQ4QDRITEBkS'\r\n" +
                        "     +'CxsSEhkWDhYYFQ0aJhkaGBweGyccGh8hHiIkIiMmGTEiHhQoPSYoJSkqKDcp'")
@@ -192,10 +216,11 @@ b +     # line3
 
   def test_slices(self):
     # Make sure we don't fail on parsing slices of the form `foo[4:]`.
-    source = "(foo.Area_Code, str(foo.Phone)[:3], str(foo.Phone)[3:], foo[:], bar[::, :])"
+    source = "(foo.Area_Code, str(foo.Phone)[:3], str(foo.Phone)[3:], foo[:], bar[::2, :], [a[:]][::-1])"
     m = self.create_mark_checker(source)
+    self.assertIn("Tuple:" + source, m.view_nodes_at(1, 0))
     self.assertEqual(m.view_nodes_at(1, 1),
-                     { "Attribute:foo.Area_Code", "Name:foo", "Tuple:"+source[1:-1] })
+                     { "Attribute:foo.Area_Code", "Name:foo" })
     self.assertEqual(m.view_nodes_at(1, 16),
                      { "Subscript:str(foo.Phone)[:3]", "Call:str(foo.Phone)", "Name:str"})
     self.assertEqual(m.view_nodes_at(1, 36),
@@ -205,7 +230,7 @@ b +     # line3
     self.assertEqual({n for n in m.view_nodes_at(1, 56) if 'Slice:' not in n},
                      { "Subscript:foo[:]", "Name:foo" })
     self.assertEqual({n for n in m.view_nodes_at(1, 64) if 'Slice:' not in n},
-                     { "Subscript:bar[::, :]", "Name:bar" })
+                     { "Subscript:bar[::2, :]", "Name:bar" })
 
   def test_adjacent_strings(self):
     source = """
@@ -217,7 +242,7 @@ bar = ('x y z'   # comment2
       )
 """
     m = self.create_mark_checker(source)
-    node_name = 'Const' if self.is_astroid_test else 'Str'
+    node_name = 'Const' if self.is_astroid_test else 'Constant'
     self.assertEqual(m.view_nodes_at(2, 6), {
       node_name + ":'x y z' \\\n'''a b c''' \"u v w\""
     })
@@ -228,20 +253,16 @@ bar = ('x y z'   # comment2
 
   def test_print_function(self):
     # This testcase imports print as function (using from __future__). Check that we can parse it.
+    # verify_all_nodes doesn't work on Python 2 because the print() call parsed in isolation
+    # is viewed as a Print node since it doesn't see the future import
     source = tools.read_fixture('astroid/nonregr.py')
-    m = self.create_mark_checker(source)
+    m = self.create_mark_checker(source, verify=six.PY3)
 
     # Line 16 is: [indent 8] print(v.get('yo'))
     self.assertEqual(m.view_nodes_at(16, 8),
                      { "Call:print(v.get('yo'))", "Expr:print(v.get('yo'))", "Name:print" })
     self.assertEqual(m.view_nodes_at(16, 14), {"Call:v.get('yo')", "Attribute:v.get", "Name:v"})
 
-    if not six.PY2:
-      # This verification fails on Py2 because to_string() doesn't know to put parens around the
-      # print function. So on Py2 we just rely on the checks above to know that it works.
-      m.verify_all_nodes(self)
-
-
   # To make sure we can handle various hard cases, we include tests for issues reported for a
   # similar project here: https://bitbucket.org/plas/thonny
 
@@ -253,7 +274,6 @@ bar = ('x y z'   # comment2
         "sünnikuupäev=str((18+int(isikukood[0:1])-1)//2)+isikukood[1:3]",
         "sünnikuupaev=str((18+int(isikukood[0:1])-1)//2)+isikukood[1:3]"):
         m = self.create_mark_checker(source)
-        m.verify_all_nodes(self)
         self.assertEqual(m.view_nodes_at(1, 0), {
           "Module:%s" % source,
           "Assign:%s" % source,
@@ -273,8 +293,7 @@ bar = ('x y z'   # comment2
         'a = f"""result: {value:{width}.{precision}}"""',
         """[f"abc {a['x']} def"]""",
         "def t():\n  return f'{function(kwarg=24)}'"):
-        m = self.create_mark_checker(source)
-        m.verify_all_nodes(self)
+        self.create_mark_checker(source)
 
     def test_adjacent_joined_strings(self):
         source = """
@@ -303,7 +322,6 @@ bar = ('x y z'   # comment2
       print_all(*arr)
     """)
     m = self.create_mark_checker(source)
-    m.verify_all_nodes(self)
     self.assertEqual(m.view_nodes_at(5, 0),
         { "Expr:print_all(*arr)", "Call:print_all(*arr)", "Name:print_all" })
     if not six.PY2 or self.is_astroid_test:
@@ -315,7 +333,6 @@ bar = ('x y z'   # comment2
     # See https://bitbucket.org/plas/thonny/issues/123/attribute-access-on-parenthesized
     source = "(x).foo()"
     m = self.create_mark_checker(source)
-    m.verify_all_nodes(self)
     self.assertEqual(m.view_nodes_at(1, 1), {"Name:x"})
     self.assertEqual(m.view_nodes_at(1, 0),
                      {"Module:(x).foo()", "Expr:(x).foo()", "Call:(x).foo()", "Attribute:(x).foo"})
@@ -324,11 +341,10 @@ bar = ('x y z'   # comment2
     # See https://bitbucket.org/plas/thonny/issues/108/ast-marker-crashes-with-conditional
     source = "a = True if True else False\nprint(a)"
     m = self.create_mark_checker(source)
-    m.verify_all_nodes(self)
     name_a = 'AssignName:a' if self.is_astroid_test else 'Name:a'
     const_true = ('Const:True' if self.is_astroid_test else
                   'Name:True' if six.PY2 else
-                  'NameConstant:True')
+                  'Constant:True')
     self.assertEqual(m.view_nodes_at(1, 0),
                      {name_a, "Assign:a = True if True else False", "Module:" + source})
     self.assertEqual(m.view_nodes_at(1, 4),
@@ -342,7 +358,6 @@ bar = ('x y z'   # comment2
     # See https://bitbucket.org/plas/thonny/issues/96/calling-lambdas-crash-the-debugger
     source = "y = (lambda x: x + 1)(2)"
     m = self.create_mark_checker(source)
-    m.verify_all_nodes(self)
     self.assertEqual(m.view_nodes_at(1, 4), {'Call:(lambda x: x + 1)(2)'})
     self.assertEqual(m.view_nodes_at(1, 15), {'BinOp:x + 1', 'Name:x'})
     if self.is_astroid_test:
@@ -358,8 +373,7 @@ bar = ('x y z'   # comment2
       "{(key, val) for key, val in ast.iter_fields(node)}",
       "{key: val for key, val in ast.iter_fields(node)}",
       "[[c for c in key] for key, val in ast.iter_fields(node)]"):
-      m = self.create_mark_checker(source)
-      m.verify_all_nodes(self)
+      self.create_mark_checker(source)
 
   def test_trailing_commas(self):
     # Make sure we handle trailing commas on comma-separated structures (e.g. tuples, sets, etc.)
@@ -368,19 +382,47 @@ bar = ('x y z'   # comment2
       "[c,d,]",
       "{e,f,}",
       "{h:1,i:2,}"):
-      m = self.create_mark_checker(source)
-      m.verify_all_nodes(self)
+      self.create_mark_checker(source)
+
+  def test_tuples(self):
+    def get_tuples(code):
+      m = self.create_mark_checker(code)
+      return [m.atok.get_text(n) for n in m.all_nodes if n.__class__.__name__ == "Tuple"]
+
+    self.assertEqual(get_tuples("a,"), ["a,"])
+    self.assertEqual(get_tuples("(a,)"), ["(a,)"])
+    self.assertEqual(get_tuples("(a),"), ["(a),"])
+    self.assertEqual(get_tuples("((a),)"), ["((a),)"])
+    self.assertEqual(get_tuples("(a,),"), ["(a,),", "(a,)"])
+    self.assertEqual(get_tuples("((a,),)"), ["((a,),)", "(a,)"])
+    self.assertEqual(get_tuples("()"), ["()"])
+    self.assertEqual(get_tuples("(),"), ["(),", "()"])
+    self.assertEqual(get_tuples("((),)"), ["((),)", "()"])
+    self.assertEqual(get_tuples("((),(a,))"), ["((),(a,))", "()", "(a,)"])
+    self.assertEqual(get_tuples("((),(a,),)"), ["((),(a,),)", "()", "(a,)"])
+    self.assertEqual(get_tuples("((),(a,),),"), ["((),(a,),),", "((),(a,),)", "()", "(a,)"])
+    self.assertEqual(get_tuples('((foo, bar),)'), ['((foo, bar),)', '(foo, bar)'])
+    self.assertEqual(get_tuples('(foo, bar),'), ['(foo, bar),', '(foo, bar)'])
+    self.assertEqual(get_tuples('def foo(a=()): ((x, (y,)),) = ((), (a,),),'), [
+      '()', '((x, (y,)),)', '(x, (y,))', '(y,)', '((), (a,),),', '((), (a,),)', '()', '(a,)'])
+    self.assertEqual(get_tuples('def foo(a=()): ((x, (y,)),) = [(), [a,],],'), [
+      '()', '((x, (y,)),)', '(x, (y,))', '(y,)', '[(), [a,],],', '()'])
+
+  def test_dict_order(self):
+    # Make sure we iterate over dict keys/values in source order.
+    # See https://github.com/gristlabs/asttokens/issues/31
+    source = 'f({1: (2), 3: 4}, object())'
+    self.create_mark_checker(source)
 
   def test_del_dict(self):
     # See https://bitbucket.org/plas/thonny/issues/24/try-del-from-dictionary-in-debugging-mode
     source = "x = {4:5}\ndel x[4]"
     m = self.create_mark_checker(source)
-    m.verify_all_nodes(self)
     self.assertEqual(m.view_nodes_at(1, 4), {'Dict:{4:5}'})
     if self.is_astroid_test:
       self.assertEqual(m.view_nodes_at(1, 5), {'Const:4'})
     else:
-      self.assertEqual(m.view_nodes_at(1, 5), {'Num:4'})
+      self.assertEqual(m.view_nodes_at(1, 5), {'Constant:4'})
     self.assertEqual(m.view_nodes_at(2, 0), {'Delete:del x[4]'})
     self.assertEqual(m.view_nodes_at(2, 4), {'Name:x', 'Subscript:x[4]'})
 
@@ -392,7 +434,6 @@ bar = ('x y z'   # comment2
           return x + y
       """)
       m = self.create_mark_checker(source)
-      m.verify_all_nodes(self)
       self.assertEqual(m.view_nodes_at(2, 0),
         {'FunctionDef:def liida_arvud(x: int, y: int) -> int:\n  return x + y'})
       if self.is_astroid_test:
@@ -407,7 +448,6 @@ bar = ('x y z'   # comment2
     # See https://bitbucket.org/plas/thonny/issues/52/range-marker-fails-with-ridastrip-split
     source = "f(x=1)\ng(a=(x),b=[y])"
     m = self.create_mark_checker(source)
-    m.verify_all_nodes(self)
     self.assertEqual(m.view_nodes_at(1, 0),
                      {'Name:f', 'Call:f(x=1)', 'Expr:f(x=1)', 'Module:' + source})
     self.assertEqual(m.view_nodes_at(2, 0),
@@ -420,7 +460,7 @@ bar = ('x y z'   # comment2
       self.assertEqual(m.view_nodes_at(2, 8), {'Keyword:b=[y]'})
     else:
       self.assertEqual(m.view_nodes_at(1, 2), {'keyword:x=1'})
-      self.assertEqual(m.view_nodes_at(1, 4), {'Num:1'})
+      self.assertEqual(m.view_nodes_at(1, 4), {'Constant:1'})
       self.assertEqual(m.view_nodes_at(2, 2), {'keyword:a=(x)'})
       self.assertEqual(m.view_nodes_at(2, 8), {'keyword:b=[y]'})
 
@@ -433,9 +473,12 @@ bar = ('x y z'   # comment2
       @deco2(a=1)
       def g(x):
         pass
+
+      @deco3()
+      def g(x):
+        pass
     """)
     m = self.create_mark_checker(source)
-    m.verify_all_nodes(self)
     # The `arguments` node has bogus positions here (and whenever there are no arguments). We
     # don't let that break our test because it's unclear if it matters to anything anyway.
     self.assertIn('FunctionDef:@deco1\ndef f():\n  pass', m.view_nodes_at(2, 0))
@@ -449,6 +492,8 @@ bar = ('x y z'   # comment2
       self.assertEqual(m.view_nodes_at(5, 0), {'FunctionDef:@deco2(a=1)\ndef g(x):\n  pass'})
     self.assertEqual(m.view_nodes_at(5, 1), {'Name:deco2', 'Call:deco2(a=1)'})
 
+    self.assertEqual(m.view_nodes_at(9, 1), {'Name:deco3', 'Call:deco3()'})
+
   def test_with(self):
     source = "with foo: pass"
     m = self.create_mark_checker(source)
@@ -466,7 +511,8 @@ bar = ('x y z'   # comment2
           with B() as b, C() as c: log(b, c)
         log(x)
       ''')
-    m = self.create_mark_checker(source)
+    # verification fails on Python2 which turns `with X, Y` turns into `with X: with Y`.
+    m = self.create_mark_checker(source, verify=six.PY3)
     self.assertEqual(m.view_nodes_at(5, 4), {
       'With:with B() as b, C() as c: log(b, c)'
     })
@@ -478,6 +524,176 @@ bar = ('x y z'   # comment2
       'With:with B() as b, C() as c: log(b, c)',
       'With:  with A() as a:\n    log(a)\n    with B() as b, C() as c: log(b, c)',
     })
-    if not six.PY2:
-      # This verification fails on Python2 which turns `with X, Y` turns into `with X: with Y`.
-      m.verify_all_nodes(self)
+
+  def test_one_line_if_elif(self):
+    source = """
+if 1: a
+elif 2: b
+    """
+    self.create_mark_checker(source)
+
+  def test_complex_numbers(self):
+    source = """
+1
+-1
+j  # not a complex number, just a name
+1j
+-1j
+1+2j
+3-4j
+1j-1j-1j-1j
+    """
+    self.create_mark_checker(source)
+
+  def test_parens_around_func(self):
+    source = textwrap.dedent(
+      '''
+      foo()
+      (foo)()
+      (lambda: 0)()
+      (lambda: ())()
+      (foo)((1))
+      (lambda: ())((2))
+      x = (obj.attribute.get_callback() or default_callback)()
+      ''')
+    m = self.create_mark_checker(source)
+    self.assertEqual(m.view_nodes_at(2, 0), {"Name:foo", "Expr:foo()", "Call:foo()"})
+    self.assertEqual(m.view_nodes_at(3, 1), {"Name:foo"})
+    self.assertEqual(m.view_nodes_at(3, 0), {"Expr:(foo)()", "Call:(foo)()"})
+    self.assertEqual(m.view_nodes_at(4, 0), {"Expr:(lambda: 0)()", "Call:(lambda: 0)()"})
+    self.assertEqual(m.view_nodes_at(5, 0), {"Expr:(lambda: ())()", "Call:(lambda: ())()"})
+    self.assertEqual(m.view_nodes_at(6, 0), {"Expr:(foo)((1))", "Call:(foo)((1))"})
+    self.assertEqual(m.view_nodes_at(7, 0), {"Expr:(lambda: ())((2))", "Call:(lambda: ())((2))"})
+    self.assertEqual(m.view_nodes_at(8, 4),
+                     {"Call:(obj.attribute.get_callback() or default_callback)()"})
+    self.assertIn('BoolOp:obj.attribute.get_callback() or default_callback', m.view_nodes_at(8, 5))
+
+  def test_complex_slice_and_parens(self):
+    source = 'f((x)[:, 0])'
+    self.create_mark_checker(source)
+
+  if six.PY3:
+    def test_sys_modules(self):
+      """
+      Verify all nodes on source files obtained from sys.modules.
+      This can take a long time as there are many modules,
+      so it only tests all modules if the environment variable
+      ASTTOKENS_SLOW_TESTS has been set.
+      """
+      modules = list(sys.modules.values())
+      if not os.environ.get('ASTTOKENS_SLOW_TESTS'):
+        modules = modules[:20]
+
+      start = time()
+      for module in modules:
+        # Don't let this test (which runs twice) take longer than 13 minutes
+        # to avoid the travis build time limit of 30 minutes
+        if time() - start > 13 * 60:
+          break
+
+        try:
+          filename = inspect.getsourcefile(module)
+        except TypeError:
+          continue
+
+        if not filename:
+          continue
+
+        filename = os.path.abspath(filename)
+        print(filename)
+        try:
+          with io.open(filename) as f:
+            source = f.read()
+        except OSError:
+          continue
+
+        # Astroid fails with a syntax error if a type comment is on its own line
+        if self.is_astroid_test and re.search(r'^\s*# type: ', source, re.MULTILINE):
+          print('Skipping', filename)
+          continue
+
+        self.create_mark_checker(source)
+
+  if six.PY3:
+    def test_dict_merge(self):
+      self.create_mark_checker("{**{}}")
+
+  def parse_snippet(self, text, node):
+    """
+    Returns the parsed AST tree for the given text, handling issues with indentation and newlines
+    when text is really an extracted part of larger code.
+    """
+    # If text is indented, it's a statement, and we need to put in a scope for indents to be valid
+    # (using textwrap.dedent is insufficient because some lines may not indented, e.g. comments or
+    # multiline strings). If text is an expression but has newlines, we parenthesize it to make it
+    # parsable.
+    # For expressions and statements, we add a dummy statement '_' before it because if it's just a
+    # string contained in an astroid.Const or astroid.Expr it will end up in the doc attribute and be
+    # a pain to extract for comparison
+    indented = re.match(r'^[ \t]+\S', text)
+    if indented:
+      return self.module.parse('def dummy():\n' + text).body[0].body[0]
+    if util.is_expr(node):
+      return self.module.parse('_\n(' + text + ')').body[1].value
+    if util.is_module(node):
+      return self.module.parse(text)
+    return self.module.parse('_\n' + text).body[1]
+
+  def test_assert_nodes_equal(self):
+    """
+    Checks that assert_nodes_equal actually fails when given different nodes
+    """
+
+    def check(s1, s2):
+      n1 = self.module.parse(s1)
+      n2 = self.module.parse(s2)
+      with self.assertRaises(AssertionError):
+        self.assert_nodes_equal(n1, n2)
+
+    check('a', 'b')
+    check('a*b', 'a+b')
+    check('a*b', 'b*a')
+    check('(a and b) or c', 'a and (b or c)')
+    check('a = 1', 'a = 2')
+    check('a = 1', 'a += 1')
+    check('a *= 1', 'a += 1')
+    check('[a for a in []]', '[a for a in ()]')
+    check("for x in y: pass", "for x in y: fail")
+    check("1", "1.0")
+    check("foo(a, b, *d, c=2, **e)",
+          "foo(a, b, *d, c=2.0, **e)")
+    check("foo(a, b, *d, c=2, **e)",
+          "foo(a, b, *d, c=2)")
+    check('def foo():\n    """xxx"""\n    None',
+          'def foo():\n    """xx"""\n    None')
+
+  nodes_classes = ast.AST
+  context_classes = [ast.expr_context]
+  iter_fields = staticmethod(ast.iter_fields)
+
+  def assert_nodes_equal(self, t1, t2):
+    # Ignore the context of each node which can change when parsing
+    # substrings of source code. We just want equal structure and contents.
+    for context_classes_group in self.context_classes:
+      if isinstance(t1, context_classes_group):
+        self.assertIsInstance(t2, context_classes_group)
+        break
+    else:
+      self.assertEqual(type(t1), type(t2))
+
+    if isinstance(t1, (list, tuple)):
+      self.assertEqual(len(t1), len(t2))
+      for vc1, vc2 in zip(t1, t2):
+        self.assert_nodes_equal(vc1, vc2)
+    elif isinstance(t1, self.nodes_classes):
+      self.assert_nodes_equal(
+        list(self.iter_fields(t1)),
+        list(self.iter_fields(t2)),
+      )
+    else:
+      # Weird bug in astroid that collapses spaces in docstrings sometimes maybe
+      if self.is_astroid_test and isinstance(t1, six.string_types):
+        t1 = re.sub(r'^ +$', '', t1, flags=re.MULTILINE)
+        t2 = re.sub(r'^ +$', '', t2, flags=re.MULTILINE)
+
+      self.assertEqual(t1, t2)
